from django.db import models
from django.db.models import Sum, Max, Min, Count, Avg
from django.db.models.functions import TruncMonth, TruncQuarter, TruncYear, TruncWeek, TruncDay
from django_datatables.columns import CurrencyColumn, CurrencyPenceColumn, ColumnLink

from advanced_report_builder.columns import ColourColumn

NUMBER_FIELDS = (models.IntegerField,
                 models.PositiveSmallIntegerField,
                 models.PositiveIntegerField,
                 models.FloatField)
DATE_FIELDS = (models.DateTimeField, models.DateField)
BOOLEAN_FIELD = models.BooleanField
CURRENCY_COLUMNS = (CurrencyColumn, CurrencyPenceColumn)
LINK_COLUMNS = ColumnLink
COLOUR_COLUMNS = ColourColumn

ANNOTATION_VALUE_YEAR = 1
ANNOTATION_VALUE_QUARTER = 2
ANNOTATION_VALUE_MONTH = 3
ANNOTATION_VALUE_WEEK = 4
ANNOTATION_VALUE_DAY = 5

ANNOTATION_VALUE_CHOICES = [(ANNOTATION_VALUE_YEAR, 'Year'),
                            (ANNOTATION_VALUE_QUARTER, 'Quarter'),
                            (ANNOTATION_VALUE_MONTH, 'Month'),
                            (ANNOTATION_VALUE_WEEK, 'Week'),
                            (ANNOTATION_VALUE_DAY, 'Day')]

ANNOTATION_VALUE_FUNCTIONS = {ANNOTATION_VALUE_YEAR: TruncYear,
                              ANNOTATION_VALUE_QUARTER: TruncQuarter,
                              ANNOTATION_VALUE_MONTH: TruncMonth,
                              ANNOTATION_VALUE_WEEK: TruncWeek,
                              ANNOTATION_VALUE_DAY: TruncDay}

ANNOTATION_CHART_SCALE = {ANNOTATION_VALUE_YEAR: 'year',
                          ANNOTATION_VALUE_QUARTER: 'quarter',
                          ANNOTATION_VALUE_MONTH: 'month',
                          ANNOTATION_VALUE_WEEK: 'week',
                          ANNOTATION_VALUE_DAY: 'day'}

GENERATE_SERIES_INTERVALS = {ANNOTATION_VALUE_YEAR: '1 year',
                             ANNOTATION_VALUE_QUARTER: '3 month',
                             ANNOTATION_VALUE_MONTH: '1 month',
                             ANNOTATION_VALUE_WEEK: '1 week',
                             ANNOTATION_VALUE_DAY: '1 day'}

ANNOTATION_CHOICE_SUM = 1
ANNOTATION_CHOICE_MAXIMUM = 2
ANNOTATION_CHOICE_MINIMUM = 3
ANNOTATION_CHOICE_COUNT = 4
ANNOTATION_CHOICE_AVERAGE_SUM_FROM_COUNT = 5

ANNOTATIONS_CHOICES = [(ANNOTATION_CHOICE_SUM, 'Sum'),
                       (ANNOTATION_CHOICE_MAXIMUM, 'Maximum'),
                       (ANNOTATION_CHOICE_MINIMUM, 'Minimum'),
                       (ANNOTATION_CHOICE_COUNT, 'Count'),
                       (ANNOTATION_CHOICE_AVERAGE_SUM_FROM_COUNT, 'Average Sum from Count')]

ANNOTATION_FUNCTIONS = {ANNOTATION_CHOICE_SUM: Sum,
                        ANNOTATION_CHOICE_MAXIMUM: Max,
                        ANNOTATION_CHOICE_MINIMUM: Min,
                        ANNOTATION_CHOICE_COUNT: Count,
                        ANNOTATION_CHOICE_AVERAGE_SUM_FROM_COUNT: Avg}

ALIGNMENT_CHOICE_LEFT = 0
ALIGNMENT_CHOICE_CENTRE = 1
ALIGNMENT_CHOICE_RIGHT = 2

ALIGNMENT_CHOICES = [(ALIGNMENT_CHOICE_LEFT, 'Left'),
                     (ALIGNMENT_CHOICE_CENTRE, 'Centre'),
                     (ALIGNMENT_CHOICE_RIGHT, 'Right')]

ALIGNMENT_CLASS = {ALIGNMENT_CHOICE_LEFT: '',
                   ALIGNMENT_CHOICE_CENTRE: 'dt-center',
                   ALIGNMENT_CHOICE_RIGHT: 'dt-right'}

DATE_FORMAT_TYPE_DD_MM_YY_SLASH = 1
DATE_FORMAT_TYPE_DD_MM_YYYY_SLASH = 2
DATE_FORMAT_TYPE_MM_DD_YY_SLASH = 3
DATE_FORMAT_TYPE_MM_DD_YYYY_SLASH = 4
DATE_FORMAT_TYPE_DD_MM_YY_DASH = 5
DATE_FORMAT_TYPE_DD_MM_YYYY_DASH = 6
DATE_FORMAT_TYPE_MM_DD_YY_DASH = 7
DATE_FORMAT_TYPE_MM_DD_YYYY_DASH = 8
DATE_FORMAT_TYPE_WORDS_MM_DD_YYYY = 9
DATE_FORMAT_TYPE_WORDS_DD_MM_YYYY = 10
DATE_FORMAT_TYPE_SHORT_WORDS_MM_DD_YYYY = 11
DATE_FORMAT_TYPE_SHORT_WORDS_DD_MM_YYYY = 12
DATE_FORMAT_TYPE_YYYY = 13
DATE_FORMAT_TYPE_YY = 14
DATE_FORMAT_TYPE_MM_YY = 15
DATE_FORMAT_TYPE_MM_YYYY = 16
DATE_FORMAT_TYPE_WORDS_MM_YY = 17
DATE_FORMAT_TYPE_WORDS_MM_YYYY = 18
DATE_FORMAT_TYPE_SHORT_WORDS_MM_YY = 19
DATE_FORMAT_TYPE_SHORT_WORDS_MM_YYYY = 20
DATE_FORMAT_TYPE_MM = 21
DATE_FORMAT_TYPE_WORDS_MM = 22
DATE_FORMAT_TYPE_SHORT_WORDS_MM = 23
DATE_FORMAT_TYPE_WW = 24
DATE_FORMAT_TYPE_WW_YYYY_DASH = 25
DATE_FORMAT_TYPE_WW_YY_DASH = 26
DATE_FORMAT_TYPE_YYYY_MM = 27


DATE_FORMAT_TYPES_DJANGO_FORMAT = {DATE_FORMAT_TYPE_DD_MM_YY_SLASH: '%d/%m/%y',
                                   DATE_FORMAT_TYPE_DD_MM_YYYY_SLASH: '%d/%m/%Y',
                                   DATE_FORMAT_TYPE_MM_DD_YY_SLASH: '%m/%d/%y',
                                   DATE_FORMAT_TYPE_MM_DD_YYYY_SLASH: '%m/%d/%Y',
                                   DATE_FORMAT_TYPE_DD_MM_YY_DASH: '%d-%m-%y',
                                   DATE_FORMAT_TYPE_DD_MM_YYYY_DASH: '%d-%m-%Y',
                                   DATE_FORMAT_TYPE_MM_DD_YY_DASH: '%m-%d-%y',
                                   DATE_FORMAT_TYPE_MM_DD_YYYY_DASH: '%m-%d-%Y',
                                   DATE_FORMAT_TYPE_WORDS_MM_DD_YYYY: '%B %d %Y',
                                   DATE_FORMAT_TYPE_WORDS_DD_MM_YYYY: '%d %B %Y',
                                   DATE_FORMAT_TYPE_SHORT_WORDS_MM_DD_YYYY: '%b %d %Y',
                                   DATE_FORMAT_TYPE_SHORT_WORDS_DD_MM_YYYY: '%d %b %Y',
                                   DATE_FORMAT_TYPE_YYYY: '%Y',
                                   DATE_FORMAT_TYPE_YY: '%y',
                                   DATE_FORMAT_TYPE_MM_YY: '%m %y',
                                   DATE_FORMAT_TYPE_YYYY_MM: '%Y %m',
                                   DATE_FORMAT_TYPE_MM_YYYY: '%m %Y',
                                   DATE_FORMAT_TYPE_WORDS_MM_YY: '%B %y',
                                   DATE_FORMAT_TYPE_WORDS_MM_YYYY: '%B %Y',
                                   DATE_FORMAT_TYPE_SHORT_WORDS_MM_YY: '%b %y',
                                   DATE_FORMAT_TYPE_SHORT_WORDS_MM_YYYY: '%b %Y',
                                   DATE_FORMAT_TYPE_MM: '%m',
                                   DATE_FORMAT_TYPE_WORDS_MM: '%B',
                                   DATE_FORMAT_TYPE_SHORT_WORDS_MM: '%b',
                                   DATE_FORMAT_TYPE_WW: '%W',
                                   DATE_FORMAT_TYPE_WW_YYYY_DASH: '%W-%Y',
                                   DATE_FORMAT_TYPE_WW_YY_DASH: '%W-%y',

                                   }

DATE_FORMAT_TYPES = [
    (DATE_FORMAT_TYPE_DD_MM_YY_SLASH, 'dd/mm/yy'),
    (DATE_FORMAT_TYPE_DD_MM_YYYY_SLASH, 'dd/mm/yyyy'),
    (DATE_FORMAT_TYPE_MM_DD_YY_SLASH, 'mm/dd/yy'),
    (DATE_FORMAT_TYPE_MM_DD_YYYY_SLASH, 'mm/dd/yyyy'),
    (DATE_FORMAT_TYPE_DD_MM_YY_DASH, 'dd-mm-yy'),
    (DATE_FORMAT_TYPE_DD_MM_YYYY_DASH, 'dd-mm-yyyy'),
    (DATE_FORMAT_TYPE_MM_DD_YY_DASH, 'mm-dd-yy'),
    (DATE_FORMAT_TYPE_MM_DD_YYYY_DASH, 'mm-dd-yyyy'),
    (DATE_FORMAT_TYPE_WORDS_MM_DD_YYYY, 'MM d yyyy'),
    (DATE_FORMAT_TYPE_WORDS_DD_MM_YYYY, 'd MM yyyy'),
    (DATE_FORMAT_TYPE_SHORT_WORDS_MM_DD_YYYY, 'M d yyyy'),
    (DATE_FORMAT_TYPE_SHORT_WORDS_DD_MM_YYYY, 'd M yyyy'),
    (DATE_FORMAT_TYPE_YYYY, 'yyyy'),
    (DATE_FORMAT_TYPE_YY, 'yy'),
    (DATE_FORMAT_TYPE_MM_YY, 'mm yy'),
    (DATE_FORMAT_TYPE_MM_YYYY, 'mm yyyy'),
    (DATE_FORMAT_TYPE_YYYY_MM, 'yyyy mm'),
    (DATE_FORMAT_TYPE_WORDS_MM_YY, 'MM yy'),
    (DATE_FORMAT_TYPE_WORDS_MM_YYYY, 'MM yyyy'),
    (DATE_FORMAT_TYPE_SHORT_WORDS_MM_YY, 'b yy'),
    (DATE_FORMAT_TYPE_SHORT_WORDS_MM_YYYY, 'b yyyy'),
    (DATE_FORMAT_TYPE_MM, 'mm'),
    (DATE_FORMAT_TYPE_WORDS_MM, 'MM'),
    (DATE_FORMAT_TYPE_SHORT_WORDS_MM, 'd'),
    (DATE_FORMAT_TYPE_WW, 'WW'),
    (DATE_FORMAT_TYPE_WW_YYYY_DASH, 'WW-YYYY'),
    (DATE_FORMAT_TYPE_WW_YY_DASH, 'WW-YY'),
]

DISPLAY_OPTION_NONE = 0
DISPLAY_OPTION_1_PER_ROW = 1
DISPLAY_OPTION_2_PER_ROW = 2
DISPLAY_OPTION_3_PER_ROW = 3
DISPLAY_OPTION_4_PER_ROW = 4


DISPLAY_OPTION_CHOICES = [
    (DISPLAY_OPTION_NONE, 'None/Default'),
    (DISPLAY_OPTION_1_PER_ROW, '1 Report per Row'),
    (DISPLAY_OPTION_2_PER_ROW, '2 Reports per Row'),
    (DISPLAY_OPTION_3_PER_ROW, '3 Reports per Row'),
    (DISPLAY_OPTION_4_PER_ROW, '4 Reports per Row'),
]

DISPLAY_OPTION_CLASSES = {DISPLAY_OPTION_1_PER_ROW: ' col-12',
                          DISPLAY_OPTION_2_PER_ROW: ' col-12 col-md-12 col-lg-6',
                          DISPLAY_OPTION_3_PER_ROW: ' col-12 col-md-12 col-lg-4',
                          DISPLAY_OPTION_4_PER_ROW: ' col-12 col-md-12 col-lg-3',
                          }

DEFAULT_DATE_FORMAT = {ANNOTATION_VALUE_YEAR: DATE_FORMAT_TYPE_YYYY,
                       ANNOTATION_VALUE_QUARTER: DATE_FORMAT_TYPE_WORDS_MM_YY,
                       ANNOTATION_VALUE_MONTH: DATE_FORMAT_TYPE_WORDS_MM_YY,
                       ANNOTATION_VALUE_WEEK: DATE_FORMAT_TYPE_WW_YY_DASH,
                       ANNOTATION_VALUE_DAY: DATE_FORMAT_TYPE_DD_MM_YY_SLASH}



from django.contrib import admin
from advanced_report_builder.models import Report, ReportType, TableReport, ReportQuery, SingleValueReport, Dashboard, \
    DashboardReport, BarChartReport, LineChartReport, PieChartReport, FunnelChartReport, ReportTag, KanbanReport, \
    KanbanReportLane, Target, CustomReport, KanbanReportDescription


@admin.register(ReportQuery)
class ReportQueryAdmin(admin.ModelAdmin):
    list_display = ('report',
                    'name',
                    'query')


class ReportQueryInline(admin.TabularInline):
    model = ReportQuery


@admin.register(Report)
class ReportAdmin(admin.ModelAdmin):
    list_display = ('name',
                    'instance_type',
                    'slug',
                    'slug_alias',
                    'version')
    exclude = ('slug',
               )

    inlines = [ReportQueryInline]


@admin.register(ReportType)
class ReportTypeAdmin(admin.ModelAdmin):
    list_display = ('name',
                    'content_type',
                    'report_builder_class_name')


@admin.register(TableReport)
class TableReportAdmin(admin.ModelAdmin):
    list_display = ('name',
                    'slug',
                    )
    exclude = ('instance_type',
               'slug',
               )
    inlines = [ReportQueryInline]


@admin.register(SingleValueReport)
class SingleValueReportAdmin(admin.ModelAdmin):
    list_display = ('name',
                    'slug',
                    )
    exclude = ('instance_type',
               'slug',
               )
    inlines = [ReportQueryInline]


@admin.register(BarChartReport)
class BarChartReportAdmin(admin.ModelAdmin):
    list_display = ('name',
                    'slug',
                    )
    exclude = ('instance_type',
               'slug',
               )
    inlines = [ReportQueryInline]


@admin.register(LineChartReport)
class LineChartReportAdmin(admin.ModelAdmin):
    list_display = ('name',
                    'slug',
                    )
    exclude = ('instance_type',
               'slug',
               )
    inlines = [ReportQueryInline]


@admin.register(PieChartReport)
class PieChartReportAdmin(admin.ModelAdmin):
    list_display = ('name',
                    'slug',
                    )
    exclude = ('instance_type',
               'slug',
               )
    inlines = [ReportQueryInline]


@admin.register(FunnelChartReport)
class FunnelChartReportAdmin(admin.ModelAdmin):
    list_display = ('name',
                    'slug',
                    )
    exclude = ('instance_type',
               'slug',
               )
    inlines = [ReportQueryInline]


@admin.register(CustomReport)
class CustomReportAdmin(admin.ModelAdmin):
    list_display = ('name',
                    'slug',
                    'output_type',
                    'view_name',
                    'settings')
    exclude = ('instance_type',
               'slug',
               'report_type')


class KanbanReportDescriptionInline(admin.TabularInline):
    model = KanbanReportDescription


class KanbanReportLaneInline(admin.TabularInline):
    model = KanbanReportLane


@admin.register(KanbanReport)
class KanbanReportAdmin(admin.ModelAdmin):
    list_display = ('name',
                    'slug',

                    )
    exclude = ('instance_type',
               'slug',
               'report_type'
               )
    inlines = [KanbanReportLaneInline,
               KanbanReportDescriptionInline]


class DashboardReportInline(admin.TabularInline):
    model = DashboardReport


@admin.register(Dashboard)
class DashboardAdmin(admin.ModelAdmin):
    list_display = ('slug',
                    'name',
                    )

    exclude = ['slug',
               ]

    inlines = [DashboardReportInline]


@admin.register(ReportTag)
class ReportTagAdmin(admin.ModelAdmin):
    list_display = ('name', 'order')


@admin.register(Target)
class TargetAdmin(admin.ModelAdmin):
    list_display = ('pk', 'name', 'target_type', 'default_value')



from crispy_forms.bootstrap import StrictButton
from django_modals.forms import ModelCrispyForm, CrispyForm


class QueryBuilderModelForm(ModelCrispyForm):
    def submit_button(self, css_class='btn-success modal-submit', button_text='Submit', **kwargs):
        return StrictButton(button_text, onclick=f'save_modal_{ self.form_id }()', css_class=css_class, **kwargs)


class QueryBuilderForm(CrispyForm):
    def submit_button(self, css_class='btn-success modal-submit', button_text='Submit', **kwargs):
        return StrictButton(button_text, onclick=f'save_modal_{ self.form_id }()', css_class=css_class, **kwargs)



from crispy_forms.layout import Div, HTML
from django.conf import settings
from django.forms import CharField
from django.urls import reverse
from django_datatables.columns import MenuColumn
from django_datatables.widgets import DataTableReorderWidget
from django_menus.menu import MenuItem, HtmlMenu

from advanced_report_builder.models import ReportQuery
from advanced_report_builder.utils import get_query_js


class MultiQueryModalMixin:
    show_order_by = True

    ajax_commands = ['datatable', 'button', 'ajax']

    def datatable_sort(self, **kwargs):
        form = self.get_form()
        widget = form.fields[kwargs['table_id'][3:]].widget
        _model = widget.attrs['table_model']
        current_sort = dict(_model.objects.filter(report_id=self.object.id).values_list('id', 'order'))
        for s in kwargs['sort']:
            if current_sort[s[1]] != s[0]:
                o = _model.objects.get(id=s[1])
                o.order = s[0]
                o.save()
        return self.command_response('null')

    def get_report_type(self, **_kwargs):
        return _kwargs['report_type']

    def button_duplicate_query(self, **_kwargs):
        query_id = _kwargs['query_id'][1:]
        report_query = ReportQuery.objects.get(pk=query_id)
        report_query_orders = report_query.reportqueryorder_set.all()
        report_query.slug = None
        report_query.pk = None
        report_query.order = None
        report_query.save()

        for report_query_order in report_query_orders:
            report_query_order.pk = None
            report_query_order.report_query = report_query
            report_query_order.save()

        report_type = self.get_report_type(**_kwargs)
        url = reverse('advanced_report_builder:query_modal',
                      kwargs={'slug': f'pk-{report_query.id}-report_type-{report_type}'})
        return self.command_response('show_modal', modal=url)

    def button_add_query(self, **_kwargs):
        report_type = self.get_report_type(**_kwargs)
        show_order_by = 1 if self.show_order_by else 0
        url = reverse('advanced_report_builder:query_modal',
                      kwargs={'slug': f'report_id-{self.object.id}-'
                                      f'report_type-{report_type}-'
                                      f'show_order_by-{show_order_by}'})
        return self.command_response('show_modal', modal=url)

    def button_edit_query(self, **_kwargs):
        query_id = _kwargs['query_id'][1:]
        report_type = self.get_report_type(**_kwargs)
        show_order_by = 1 if self.show_order_by else 0
        url = reverse('advanced_report_builder:query_modal',
                      kwargs={'slug': f'pk-{query_id}-'
                                      f'report_type-{report_type}-'
                                      f'show_order_by-{show_order_by}'})
        return self.command_response('show_modal', modal=url)

    def query_menu(self):
        edit_query_js = get_query_js('edit_query', 'query_id')
        duplicate_query_js = get_query_js('duplicate_query', 'query_id')

        description_edit_menu_items = [
            MenuItem(duplicate_query_js.replace('"', "&quot;"),
                     menu_display='Duplicate',
                     css_classes='btn btn-sm btn-outline-dark',
                     font_awesome='fas fa-clone',
                     link_type=MenuItem.HREF),

            MenuItem(edit_query_js.replace('"', "&quot;"),
                     menu_display='Edit',
                     css_classes='btn btn-sm btn-outline-dark',
                     font_awesome='fas fa-pencil',
                     link_type=MenuItem.HREF)]
        return description_edit_menu_items


    def add_extra_queries(self, form, fields):
        add_query_js = 'django_modal.process_commands_lock([{"function": "post_modal", ' \
                       '"button": {"button": "add_query"}}])'

        description_add_menu_items = [

            MenuItem(add_query_js.replace('"', "&quot;"),
                     menu_display='Add Query Version',
                     css_classes='btn btn-primary',
                     font_awesome='fas fa-pencil',
                     link_type=MenuItem.HREF)]

        menu = HtmlMenu(self.request,
                        'advanced_report_builder/datatables/onclick_menu.html').add_items(
            *description_add_menu_items)

        fields.append(Div(HTML(menu.render()), css_class='form-buttons'))
        description_edit_menu_items = self.query_menu()


        form.fields['queries'] = CharField(
            required=False,
            label='Queries Versions',
            widget=DataTableReorderWidget(
                model=ReportQuery,
                order_field='order',
                fields=['_.index',
                        '.id',
                        'name',
                        MenuColumn(column_name='menu', field='id',
                                   column_defs={'orderable': False, 'className': 'dt-right'},
                                   menu=HtmlMenu(self.request,
                                                 'advanced_report_builder/datatables/onclick_menu.html').add_items(
                                       *description_edit_menu_items
                                   )),
                        ],
                attrs={'filter': {'report__id': self.object.id}}))
        fields.append('queries')

    def post_save(self, created, form):
        if created:
            self.modal_redirect(self.request.resolver_match.view_name, slug=f'pk-{self.object.id}-new-True')
        else:
            url_name = getattr(settings, 'REPORT_BUILDER_DETAIL_URL_NAME', '')
            if url_name and self.slug.get('new'):
                url = reverse(url_name, kwargs={'slug': self.object.slug})
                self.command_response('redirect', url=url)



# Generated by Django 3.2.7 on 2022-03-10 11:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('advanced_report_builder', '0005_auto_20220303_0958'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='reportquery',
            options={'ordering': ['order'], 'verbose_name_plural': 'Report queries'},
        ),
        migrations.AddField(
            model_name='reportquery',
            name='order',
            field=models.PositiveSmallIntegerField(default=1),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='reportquery',
            name='name',
            field=models.CharField(default='Standard', max_length=256),
        ),
        migrations.AlterUniqueTogether(
            name='reportquery',
            unique_together=set(),
        ),
    ]



import copy
import json
import operator
from functools import reduce

from django.db.models import FloatField, ExpressionWrapper
from django.db.models.functions import NullIf
from django_datatables.columns import CurrencyPenceColumn, CurrencyColumn

from advanced_report_builder.columns import ReportBuilderCurrencyPenceColumn, ReportBuilderCurrencyColumn, \
    ReportBuilderNumberColumn
from advanced_report_builder.field_utils import ReportBuilderFieldUtils
from advanced_report_builder.filter_query import FilterQueryMixin
from advanced_report_builder.globals import ANNOTATION_FUNCTIONS, ANNOTATION_CHOICE_COUNT, ALIGNMENT_CLASS
from advanced_report_builder.utils import decode_attribute


class ReportUtilsMixin(ReportBuilderFieldUtils, FilterQueryMixin):
    use_annotations = True
    number_field = ReportBuilderNumberColumn

    def get_number_field(self, annotations_type, index, table_field, data_attr, fields, col_type_override,
                         extra_filter=None, title_suffix='', multiple_index=0, decimal_places=None,
                         convert_currency_fields=False, totals=None, divider=None, additional_options=None):
        field_name = table_field['field']
        alignment_class = ALIGNMENT_CLASS.get(int(data_attr.get('alignment', 0)))
        new_field_name = field_name
        css_class = None
        annotation_filter = None
        if annotations_type != 0:
            b64_filter = data_attr.get('filter')
            if b64_filter:
                _filter = decode_attribute(b64_filter)
                _filter = json.loads(_filter)
                annotation_filter = self.process_filters(search_filter_data=_filter, extra_filter=extra_filter)
            elif extra_filter:
                annotation_filter = reduce(operator.and_, [extra_filter])

        if int(data_attr.get('display_heading', 1)) == 0:
            title = ''
        elif int(data_attr.get('append_column_title', 0)) == 0:
            title = title_suffix
        else:
            title = title_suffix + ' ' + table_field.get('title')
        if col_type_override:
            col_type_override.table = None
            field = copy.deepcopy(col_type_override)
            if field.model_path and isinstance(field.field, str) and field.field.startswith(field.model_path):
                raw_field_name = field.field[len(field.model_path):]
            else:
                raw_field_name = field.field

            if convert_currency_fields:
                if isinstance(field, CurrencyPenceColumn):
                    field.__class__ = ReportBuilderCurrencyPenceColumn
                elif isinstance(field, CurrencyColumn):
                    field.__class__ = ReportBuilderCurrencyColumn

            if field.annotations:
                if not self.use_annotations:
                    field.options['calculated'] = True
                    field.aggregations = col_type_override.annotations
                    field.annotations = []
                if title:
                    field.title = title
                self.set_extra_number_field_kwargs(data_attr=data_attr,
                                                   options=field.options,
                                                   multiple_index=multiple_index,
                                                   additional_options=additional_options)

            elif annotations_type == ANNOTATION_CHOICE_COUNT:
                new_field_name = self.get_new_annotation_field_name(annotations_type=annotations_type,
                                                                    field_name=field_name,
                                                                    index=index,
                                                                    data_attr=data_attr)
                number_function_kwargs = {}
                if title:
                    number_function_kwargs['title'] = title

                if alignment_class != '':
                    css_class = alignment_class
                    number_function_kwargs['column_defs'] = {'className': css_class}

                function_type = ANNOTATION_FUNCTIONS[annotations_type]

                number_function_kwargs['options'] = {}
                self.set_extra_number_field_kwargs(data_attr=data_attr,
                                                   options=number_function_kwargs['options'],
                                                   multiple_index=multiple_index,
                                                   additional_options=additional_options)
                if annotation_filter:
                    function = function_type(raw_field_name, filter=annotation_filter)
                else:
                    function = function_type(raw_field_name)
                if divider:
                    function = ExpressionWrapper(function / NullIf(divider, 0), output_field=FloatField())
                if self.use_annotations:
                    number_function_kwargs['annotations'] = {new_field_name: function}
                else:
                    number_function_kwargs['options']['calculated'] = True
                    number_function_kwargs['aggregations'] = {new_field_name: function}
                    number_function_kwargs['annotations'] = []

                number_function_kwargs.update({'field': new_field_name,
                                               'column_name': new_field_name})
                field = self.number_field(**number_function_kwargs)
            else:
                css_class = field.column_defs.get('className')
                if css_class is None:
                    css_class = alignment_class
                else:
                    css_class += ' ' + alignment_class
                field.column_defs['className'] = css_class

                if title:
                    field.title = title
                if annotations_type != 0:
                    new_field_name = self.get_new_annotation_field_name(annotations_type=annotations_type,
                                                                        field_name=field_name,
                                                                        index=index,
                                                                        data_attr=data_attr)
                    function_type = ANNOTATION_FUNCTIONS[annotations_type]
                    if annotation_filter:
                        function = function_type(raw_field_name, filter=annotation_filter)
                    else:
                        function = function_type(raw_field_name)
                    if divider:
                        function = ExpressionWrapper(function / NullIf(divider, 0), output_field=FloatField())
                    if self.use_annotations:
                        field.annotations = {new_field_name: function}
                    else:
                        field.options['calculated'] = True
                        field.aggregations = {new_field_name: function}
                        field.annotations = []

                    field.field = new_field_name
                    self.set_extra_number_field_kwargs(data_attr=data_attr,
                                                       options=field.options,
                                                       multiple_index=multiple_index,
                                                       additional_options=additional_options)
            fields.append(field)
        else:
            number_function_kwargs = {'title': title}
            decimal_places = data_attr.get('decimal_places', decimal_places)
            if decimal_places:
                number_function_kwargs['decimal_places'] = int(decimal_places)
            if annotations_type:
                number_function_kwargs['options'] = {}
                self.set_extra_number_field_kwargs(data_attr=data_attr,
                                                   options=number_function_kwargs['options'],
                                                   multiple_index=multiple_index,
                                                   additional_options=additional_options)

                new_field_name = self.get_new_annotation_field_name(annotations_type=annotations_type,
                                                                    field_name=field_name,
                                                                    index=index,
                                                                    data_attr=data_attr)
                function_type = ANNOTATION_FUNCTIONS[annotations_type]
                if annotation_filter:
                    function = function_type(field_name, filter=annotation_filter)
                else:
                    function = function_type(field_name)

                if self.use_annotations:
                    number_function_kwargs['annotations'] = {new_field_name: function}
                else:
                    number_function_kwargs['options']['calculated'] = True
                    number_function_kwargs['aggregations'] = {new_field_name: function}
                    number_function_kwargs['annotations'] = []
                field_name = new_field_name
            number_function_kwargs.update({'field': field_name,
                                           'column_name': field_name,
                                           'model_path': ''})
            field = self.number_field(**number_function_kwargs)
            fields.append(field)

        if totals is not None:
            show_total = data_attr.get('show_totals')
            if show_total == '1':
                self.set_number_total(totals=totals,
                                      field_name=new_field_name,
                                      col_type_override=col_type_override,
                                      decimal_places=decimal_places,
                                      css_class=css_class)
        return field_name

    @staticmethod
    def get_new_annotation_field_name(annotations_type, field_name, index, data_attr):
        annotation_column_id = data_attr.get('annotation_column_id')
        if annotation_column_id:
            return decode_attribute(annotation_column_id)
        return f'{annotations_type}_{field_name}_{index}'

    def set_extra_number_field_kwargs(self, data_attr, options, multiple_index, additional_options):
        pass

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def set_number_total(self, totals, field_name, col_type_override, decimal_places, css_class=''):
        totals[field_name] = {'sum': 'to_fixed', 'decimal_places': decimal_places, 'css_class': css_class}

    def set_annotation_total(self, totals, field_name, col_type_override, decimal_places, css_class=''):
        return self.set_number_total(totals, field_name, col_type_override, decimal_places, css_class)

    def set_percentage_total(self, totals, field_name, denominator, numerator, decimal_places, css_class=''):
        totals[field_name] = {'sum': 'percentage',
                              'decimal_places': decimal_places,
                              'denominator': denominator,
                              'numerator': numerator,
                              'css_class': css_class}



from ajax_helpers.mixins import AjaxHelpers
from django_menus.menu import MenuItem, MenuMixin

from advanced_report_builder.utils import make_slug_str


class ReportBase(AjaxHelpers, MenuMixin):
    # noinspection PyUnusedLocal
    @staticmethod
    def duplicate_menu(request, report_id):
        return [MenuItem(f'advanced_report_builder:duplicate_report_modal,pk-{report_id}',
                         css_classes=['btn-success'])]

    def get_dashboard_class(self, report):
        return None

    def queries_menu(self, report, dashboard_report):
        query_slug = f'query{report.id}'
        if dashboard_report is not None:
            if not dashboard_report.show_versions:
                return []
            query_slug += f'_{dashboard_report.id}'

        report_queries = report.reportquery_set.all()
        if len(report_queries) > 1:
            dropdown = []
            for report_query in report_queries:
                slug_str = make_slug_str(self.slug, overrides={query_slug: report_query.id})
                dropdown.append((self.request.resolver_match.view_name,
                                 report_query.name, {'url_kwargs': {'slug': slug_str}}))
            return [MenuItem(menu_display='Version', no_hover=True, css_classes='btn-secondary',
                             dropdown=dropdown)]
        return []


from django.db import ProgrammingError
from django_datatables.datatables import DatatableView, DatatableError
from django_datatables.helpers import row_link
from django_menus.menu import MenuItem

from advanced_report_builder.columns import ArrowColumn
from advanced_report_builder.exceptions import ReportError
from advanced_report_builder.utils import get_report_builder_class
from advanced_report_builder.utils import split_slug
from advanced_report_builder.views.datatables.utils import TableUtilsMixin
from advanced_report_builder.views.report import ReportBase


class TableView(ReportBase, TableUtilsMixin, DatatableView):
    template_name = 'advanced_report_builder/datatables/report.html'
    menu_display = ''

    def __init__(self, *args, **kwargs):
        self.table_id = None
        super().__init__(*args, **kwargs)

    def add_tables(self):
        return None

    def dispatch(self, request, *args, **kwargs):
        self.slug = split_slug(self.kwargs['slug'])
        self.report = kwargs.get('report')
        self.table_report = self.report.tablereport
        self.dashboard_report = kwargs.get('dashboard_report')
        self.enable_edit = kwargs.get('enable_edit')
        if self.dashboard_report:
            self.table_id = f'tabledashboard_{self.dashboard_report.id}'
        else:
            self.table_id = f'table_{self.table_report.id}'

        self.base_model = self.table_report.get_base_model()
        self.add_table(self.table_id, model=self.base_model)

        try:
            return super().dispatch(request, *args, **kwargs)
        except DatatableError as de:
            raise ReportError(de.args[0])
        except ProgrammingError as de:
            raise ReportError(de.args[0])

    def setup_table(self, table):
        table.extra_filters = self.extra_filters
        base_model = self.table_report.get_base_model()
        table_fields = self.table_report.table_fields
        pivot_fields = self.table_report.pivot_fields
        fields_used = set()
        fields_map = {}
        report_builder_class = get_report_builder_class(model=base_model,
                                                        report_type=self.table_report.report_type)
        self.process_query_results(report_builder_class=report_builder_class,
                                   table=table,
                                   base_model=base_model,
                                   fields_used=fields_used,
                                   fields_map=fields_map,
                                   table_fields=table_fields,
                                   pivot_fields=pivot_fields)

        if self.table_report.order_by_field:
            order_by_field = self.table_report.order_by_field
            if order_by_field not in fields_used:
                table.add_columns(f'.{order_by_field}')
            elif order_by_field in fields_map:
                order_by_field = fields_map[order_by_field]
            if self.table_report.order_by_ascending:
                table.sort(order_by_field)
            else:
                table.sort(f'-{order_by_field}')

        table.table_options['pageLength'] = self.table_report.page_length
        table.table_options['bStateSave'] = False

        if self.table_report.has_clickable_rows and self.table_report.link_field and self.kwargs.get('enable_links'):
            table.table_classes.append('row_link')
            table.add_columns(ArrowColumn(column_name='arrow_icon'))
            _, col_type_override, _, _ = self.get_field_details(base_model=base_model,
                                                                field=self.table_report.link_field,
                                                                report_builder_class=report_builder_class)
            if isinstance(col_type_override.field, list):
                field = col_type_override.field[0]
            else:
                field = 'id'
            if field not in fields_used:
                table.add_columns(f'.{field}')
            table.table_options['row_href'] = row_link(col_type_override.url, field)

    def get_title(self):
        title = super().get_title()
        report_queries_count = self.table_report.reportquery_set.all().count()
        if report_queries_count > 1:
            version_name = self.get_report_query(report=self.table_report).name
            title += f' ({version_name})'
        return title

    def add_to_context(self, **kwargs):
        return {'title': self.get_title(),
                'table_report': self.table_report}

    def setup_menu(self):
        super().setup_menu()
        if self.dashboard_report and self.enable_edit:
            report_menu = self.pod_dashboard_edit_menu()
        elif self.dashboard_report and not self.enable_edit:
            report_menu = self.pod_dashboard_view_menu()
        else:
            report_menu = self.pod_report_menu()

        self.add_menu('button_menu', 'button_group').add_items(
            *report_menu,
            *self.queries_menu(report=self.report, dashboard_report=self.dashboard_report),
        )

    def pod_dashboard_edit_menu(self):
        return [MenuItem(f'advanced_report_builder:dashboard_report_modal,pk-{self.dashboard_report.id}',
                         menu_display='Edit',
                         font_awesome='fas fa-pencil-alt', css_classes=['btn-primary'])]

    # noinspection PyMethodMayBeStatic
    def pod_dashboard_view_menu(self):
        return []

    def pod_report_menu(self):

        query_id = self.slug.get(f'query{self.table_report.id}')
        slug_str = ''
        if query_id:
            slug_str = f'-query_id-{query_id}'

        return self.edit_report_menu(request=self.request, chart_report_id=self.table_report.id, slug_str=slug_str)

    def edit_report_menu(self, request, chart_report_id, slug_str):
        return [MenuItem(f'advanced_report_builder:table_modal,pk-{chart_report_id}{slug_str}',
                         menu_display='Edit',
                         font_awesome='fas fa-pencil-alt', css_classes=['btn-primary']),
                *self.duplicate_menu(request=self.request, report_id=chart_report_id)
                ]

    def get_dashboard_class(self, report):
        pivot_fields = report.tablereport.pivot_fields
        if pivot_fields is not None and len(pivot_fields) > 0:
            return 'p-0'
        return ''



from django_datatables.columns import ColumnNameError
from django_datatables.datatables import ColumnInitialisor
from django.apps import apps
from advanced_report_builder.exceptions import ReportError
from advanced_report_builder.globals import DATE_FIELDS, NUMBER_FIELDS, LINK_COLUMNS, COLOUR_COLUMNS
from advanced_report_builder.utils import get_report_builder_class


class ReportBuilderFieldUtils:
    column_initialisor_cls = ColumnInitialisor
    def get_field_details(self, base_model, field, report_builder_class, table=None, field_attr=None):

        if field_attr is None:
            field_attr = {}
        if isinstance(field, str) and field in report_builder_class.field_classes:
            field = report_builder_class.field_classes[field]

        path = field
        original_column_initialisor = self.column_initialisor_cls(start_model=base_model,
                                                                  path=field,
                                                                  table=table,
                                                                  **field_attr)

        try:
            columns = original_column_initialisor.get_columns()
        except ColumnNameError as e:
            raise ReportError(e)
        django_field = original_column_initialisor.django_field
        col_type_override = None
        if columns:
            col_type_override = columns[0]
            if isinstance(field, str) and '__' in field and col_type_override.model_path == '':
                # I think there is a bug in datatables where the model path isn't always set! hence the following code
                field_parts = field.split('__')[:-1]
                model_path = '__'.join(field_parts) + '__'
                col_type_override.model_path = model_path
                col_type_override.field = col_type_override.field  # actions the setter
                path = model_path

        if django_field is None and columns:
            col_type_override = columns[0]
            if isinstance(col_type_override.field, str):
                if isinstance(field, str):
                    path_parts = field.split('__')[:-1]
                    path_parts.append(col_type_override.field.split('__')[-1])
                    path = '__'.join(path_parts)
                    column_initialisor = self.column_initialisor_cls(start_model=base_model, path=path, table=table,
                                                                     **field_attr)
                    column_initialisor.get_columns()
                    django_field = column_initialisor.django_field
                else:
                    column_initialisor = self.column_initialisor_cls(start_model=base_model,
                                                                     path=col_type_override.field,
                                                                     table=table,
                                                                     **field_attr)
                    column_initialisor.get_columns()
                    django_field = column_initialisor.django_field

        return django_field, col_type_override, columns, path


    def get_field_display_value(self, field_type, fields_values, base_model,
                                report_builder_class, selected_field_value, for_select2=False):
        if field_type == 'date':
            self._get_date_fields(base_model=base_model,
                                  fields=fields_values,
                                  report_builder_class=report_builder_class,
                                  selected_field_id=selected_field_value)
        elif field_type == 'link':
            self._get_column_link_fields(base_model=base_model,
                                         fields=fields_values,
                                         report_builder_class=report_builder_class,
                                         selected_field_id=selected_field_value)
        elif field_type == 'number':
            self._get_number_fields(base_model=base_model,
                                    fields=fields_values,
                                    report_builder_class=report_builder_class,
                                    selected_field_id=selected_field_value)
        elif field_type == 'colour':
            self._get_colour_fields(base_model=base_model,
                                    fields=fields_values,
                                    report_builder_class=report_builder_class,
                                    selected_field_id=selected_field_value)
        elif field_type == 'all':
            self._get_fields(base_model=base_model,
                             fields=fields_values,
                             report_builder_class=report_builder_class,
                             selected_field_id=selected_field_value,
                             for_select2=for_select2)
        elif field_type == 'order':
            self._get_fields(base_model=base_model,
                             fields=fields_values,
                             report_builder_class=report_builder_class,
                             selected_field_id=selected_field_value,
                             for_select2=for_select2,
                             show_order_by_fields=True)
        elif field_type == 'django_order':
            self._get_fields(base_model=base_model,
                             fields=fields_values,
                             report_builder_class=report_builder_class,
                             selected_field_id=selected_field_value,
                             for_select2=for_select2,
                             show_order_by_fields=True)

    def _get_date_fields(self, base_model, fields, report_builder_class, selected_field_id=None, search_string=None):
        return self._get_fields(base_model=base_model,
                                fields=fields,
                                report_builder_class=report_builder_class,
                                selected_field_id=selected_field_id,
                                field_types=DATE_FIELDS,
                                for_select2=True,
                                allow_annotations_fields=False,
                                search_string=search_string)

    def _get_number_fields(self, base_model, fields, report_builder_class, selected_field_id=None, search_string=None):
        return self._get_fields(base_model=base_model,
                                fields=fields,
                                report_builder_class=report_builder_class,
                                selected_field_id=selected_field_id,
                                field_types=NUMBER_FIELDS,
                                for_select2=True,
                                search_string=search_string
                                )

    def _get_column_link_fields(self, base_model, fields, report_builder_class,
                                selected_field_id=None, search_string=None):
        return self._get_fields(base_model=base_model,
                                fields=fields,
                                report_builder_class=report_builder_class,
                                selected_field_id=selected_field_id,
                                column_types=LINK_COLUMNS,
                                for_select2=True,
                                search_string=search_string,
                                allow_annotations_fields=False
                                )

    def _get_colour_fields(self, base_model, fields, report_builder_class,
                           selected_field_id=None, search_string=None):
        return self._get_fields(base_model=base_model,
                                fields=fields,
                                report_builder_class=report_builder_class,
                                selected_field_id=selected_field_id,
                                column_types=COLOUR_COLUMNS,
                                for_select2=True,
                                search_string=search_string,
                                allow_annotations_fields=False
                                )

    def _get_fields(self, base_model, fields, report_builder_class, tables=None,
                    prefix='', title_prefix='', title=None, colour=None,
                    previous_base_model=None, selected_field_id=None, for_select2=False,
                    pivot_fields=None, allow_annotations_fields=True, field_types=None,
                    column_types=None, search_string=None, show_order_by_fields=False, extra_fields=None,
                    must_have_django_field=False, allow_pivots=True, include_mathematical_columns=False):
        if title is None:
            title = report_builder_class.title
        if colour is None:
            colour = report_builder_class.colour

        if tables is not None:
            tables.append({'name': title,
                           'colour': colour})

        report_builder_class_fields = report_builder_class.fields

        if extra_fields:
            report_builder_class_fields += extra_fields

        for report_builder_field in report_builder_class_fields:

            if (not isinstance(report_builder_field, str) or
                    report_builder_field not in report_builder_class.exclude_display_fields or
                    (show_order_by_fields and report_builder_field in report_builder_class.order_by_fields)):
                django_field, col_type_override, columns, _ = self.get_field_details(
                    base_model=base_model,
                    field=report_builder_field,
                    report_builder_class=report_builder_class)
                if django_field is None and must_have_django_field:
                    continue
                for column in columns:
                    if ((field_types is None and column_types is None) or
                            (field_types is not None and isinstance(django_field, field_types)) or
                            (column_types is not None and isinstance(col_type_override, column_types)) or
                            (allow_annotations_fields and column.annotations)):
                        full_id = prefix + column.column_name
                        if selected_field_id is None or selected_field_id == full_id:

                            if column.title == '':
                                full_title = title_prefix + col_type_override.title_from_name(column.column_name)
                            else:
                                full_title = title_prefix + column.title
                            if self._is_search_match(search_string=search_string, title=full_title):
                                if for_select2:
                                    fields.append({'id': full_id,
                                                   'text': full_title})
                                else:
                                    fields.append({'field': full_id,
                                                   'label': full_title,
                                                   'colour': colour})

        if allow_pivots and not for_select2 and pivot_fields is not None:
            for pivot_code, pivot_field in report_builder_class.pivot_fields.items():
                full_id = prefix + pivot_code
                full_title = title_prefix + pivot_field['title']
                if self._is_search_match(search_string=search_string, title=full_title):
                    pivot_fields.append({'field': full_id,
                                         'label': title_prefix + pivot_field['title'],
                                         'colour': colour})

        for include_field, include in report_builder_class.includes.items():
            app_label, model, report_builder_fields_str = include['model'].split('.')
            local_allow_pivots = allow_pivots
            if local_allow_pivots and not include.get('allow_pivots', True):
                local_allow_pivots = False

            new_model = apps.get_model(app_label, model)
            if new_model != previous_base_model:
                new_report_builder_class = get_report_builder_class(model=new_model,
                                                                    class_name=report_builder_fields_str)
                self._get_fields(base_model=new_model,
                                 fields=fields,
                                 report_builder_class=new_report_builder_class,
                                 tables=tables,
                                 prefix=f"{prefix}{include_field}__",
                                 title_prefix=f"{title_prefix}{include['title']} -> ",
                                 title=include.get('title'),
                                 colour=include.get('colour'),
                                 previous_base_model=base_model,
                                 selected_field_id=selected_field_id,
                                 for_select2=for_select2,
                                 pivot_fields=pivot_fields,
                                 allow_annotations_fields=allow_annotations_fields,
                                 field_types=field_types,
                                 column_types=column_types,
                                 search_string=search_string,
                                 show_order_by_fields=show_order_by_fields,
                                 allow_pivots=local_allow_pivots)

        if include_mathematical_columns:
            fields.extend([{'colour': '#D4AF37', 'field': 'rb_addition', 'label': 'Maths: Addition Field'},
                          {'colour': '#D4AF37', 'field': 'rb_subtraction', 'label': 'Maths: Subtraction Field'},
                          {'colour': '#D4AF37', 'field': 'rb_times', 'label': 'Maths: Times Field'},
                          {'colour': '#D4AF37', 'field': 'rb_division', 'label': 'Maths: Division Field'},
                          {'colour': '#D4AF37', 'field': 'rb_percentage', 'label': 'Maths: Percentage Field'}]
                          )

    @staticmethod
    def _is_search_match(search_string, title):
        if search_string is None:
            return True
        return search_string.lower() in title.lower()



from advanced_report_builder.columns import RecordCountColumn


class ReportBuilderFields:
    colour = None
    title = None
    fields = []
    pivot_fields = {}
    exclude_search_fields = set()
    exclude_display_fields = set()
    order_by_fields = set()
    url = None
    includes = {}
    default_columns = []

    field_classes = {'record_count': RecordCountColumn()}

    default_multiple_column_text = ''
    default_multiple_column_fields = []
    default_multiple_pk = 'id'

    extra_chart_field = ['record_count']





from django_modals.widgets.widgets import Toggle


class RBToggle(Toggle):

    def __init__(self, attrs=None, check_test=None):

        if attrs is None:
            attrs = {'data-onstyle': 'success',
                     'data-on': 'YES',
                     'data-off': 'NO'}

        super().__init__(attrs, check_test)



import base64
import json

from crispy_forms.bootstrap import StrictButton
from crispy_forms.layout import Div
from django.forms import CharField, ChoiceField, BooleanField, IntegerField
from django.urls import reverse
from django_modals.fields import FieldEx
from django_modals.form_helpers import HorizontalNoEnterHelper
from django_modals.modals import FormModal
from django_modals.processes import PROCESS_EDIT_DELETE, PERMISSION_OFF
from django_modals.widgets.select2 import Select2Multiple
from django_modals.widgets.widgets import Toggle

from advanced_report_builder.globals import DATE_FIELDS, NUMBER_FIELDS, ANNOTATION_VALUE_CHOICES, ANNOTATIONS_CHOICES, \
    DATE_FORMAT_TYPES, CURRENCY_COLUMNS, LINK_COLUMNS, ALIGNMENT_CHOICES
from advanced_report_builder.models import TableReport, ReportType
from advanced_report_builder.toggle import RBToggle
from advanced_report_builder.utils import split_attr, encode_attribute, decode_attribute, get_report_builder_class
from advanced_report_builder.views.charts_base import ChartBaseFieldForm
from advanced_report_builder.views.modals_base import QueryBuilderModalBaseMixin, QueryBuilderModalBase
from advanced_report_builder.views.query_modal.mixin import MultiQueryModalMixin


class TableModal(MultiQueryModalMixin, QueryBuilderModalBase):
    template_name = 'advanced_report_builder/datatables/modal.html'
    size = 'xl'
    model = TableReport
    process = PROCESS_EDIT_DELETE
    permission_delete = PERMISSION_OFF
    show_order_by = False

    widgets = {'report_tags': Select2Multiple}

    form_fields = ['name',
                   'notes',
                   ('has_clickable_rows', {'widget': Toggle(attrs={'data-onstyle': 'success',
                                                                   'data-on': 'YES',
                                                                   'data-off': 'NO'})}),
                   'link_field',
                   'order_by_field',
                   ('order_by_ascending', {'widget': Toggle(attrs={'data-onstyle': 'success',
                                                                   'data-on': 'YES',
                                                                   'data-off': 'NO'})}),

                   'page_length',
                   'report_type',
                   'report_tags',
                   'table_fields',
                   'pivot_fields',
                   ]

    def form_setup(self, form, *_args, **_kwargs):
        url = reverse('advanced_report_builder:table_field_modal',
                      kwargs={'slug': 'selector-99999-data-FIELD_INFO-report_type_id-REPORT_TYPE_ID'})

        pivot_url = reverse('advanced_report_builder:table_pivot_modal',
                            kwargs={'slug': 'selector-99999-data-FIELD_INFO-report_type_id-REPORT_TYPE_ID'})

        form.add_trigger('has_clickable_rows', 'onchange', [
            {'selector': '#div_id_link_field', 'values': {'checked': 'show'}, 'default': 'hide'},
        ])

        form.fields['notes'].widget.attrs['rows'] = 3

        if 'data' in _kwargs:
            link_field = _kwargs['data'].get('link_field')
            order_by_field = _kwargs['data'].get('order_by_field')
            report_type_id = _kwargs['data'].get('report_type')
            report_type = ReportType.objects.filter(id=report_type_id).first()  # can be None
        else:
            link_field = form.instance.link_field
            order_by_field = form.instance.order_by_field
            report_type = form.instance.report_type

        self.setup_field(field_type='link',
                         form=form,
                         field_name='link_field',
                         selected_field_id=link_field,
                         report_type=report_type)

        self.setup_field(field_type='order',
                         form=form,
                         field_name='order_by_field',
                         selected_field_id=order_by_field,
                         report_type=report_type)

        fields = ['name',
                  'notes',
                  'report_type',
                  'report_tags',
                  FieldEx('has_clickable_rows', template='django_modals/fields/label_checkbox.html'),
                  'link_field',
                  'order_by_field',
                  FieldEx('order_by_ascending', template='django_modals/fields/label_checkbox.html'),
                  FieldEx('page_length', template='django_modals/fields/label_checkbox.html'),
                  FieldEx('table_fields',
                          template='advanced_report_builder/select_column.html',
                          extra_context={'select_column_url': url,
                                         'command_prefix': ''}),
                  FieldEx('pivot_fields',
                          template='advanced_report_builder/datatables/select_pivot.html',
                          extra_context={'select_column_url': pivot_url}),
                  ]

        if self.object.id:
            self.add_extra_queries(form=form, fields=fields)

        return fields

    def form_valid(self, form):
        org_id = self.object.pk if hasattr(self, 'object') else None
        save_function = getattr(form, 'save', None)
        if save_function:
            save_function()
        self.post_save(created=org_id is None, form=form)
        if not self.response_commands:
            self.add_command('reload')
        return self.command_response()

    def ajax_get_fields(self, **kwargs):
        report_type_id = kwargs['report_type']
        report_builder_class, base_model = self.get_report_builder_class(report_type_id=report_type_id)
        fields = []
        tables = []
        pivot_fields = []
        self._get_fields(base_model=base_model,
                         fields=fields,
                         tables=tables,
                         report_builder_class=report_builder_class,
                         pivot_fields=pivot_fields,
                         include_mathematical_columns=True)

        self.add_command('report_fields', data=json.dumps({'fields': fields, 'tables': tables}))
        self.add_command('report_pivots', data=json.dumps({'pivot_fields': pivot_fields}))

        return self.command_response()

    def select2_link_field(self, **kwargs):
        return self.get_fields_for_select2(field_type='link',
                                           report_type=kwargs['report_type'],
                                           search_string=kwargs.get('search'))

    def select2_order_by_field(self, **kwargs):
        return self.get_fields_for_select2(field_type='order',
                                           report_type=kwargs['report_type'],
                                           search_string=kwargs.get('search'))


class TableFieldForm(ChartBaseFieldForm):
    cancel_class = 'btn-secondary modal-cancel'

    def cancel_button(self, css_class=cancel_class, **kwargs):
        commands = [{'function': 'close'}]
        return self.button('Cancel', commands, css_class, **kwargs)

    def submit_button(self, css_class='btn-success modal-submit', button_text='Submit', **kwargs):
        if self.django_field is not None and isinstance(self.django_field, NUMBER_FIELDS):
            return StrictButton(button_text, onclick=f'save_modal_{self.form_id}()', css_class=css_class, **kwargs)
        else:
            return super().submit_button(css_class, button_text, **kwargs)

    def is_mathematical_field(self, data):
        return self.django_field is None and data['field'] in ['rb_addition',
                                                               'rb_subtraction',
                                                               'rb_times',
                                                               'rb_division',
                                                               'rb_percentage']

    def setup_modal(self, *args, **kwargs):
        data = json.loads(base64.b64decode(self.slug['data']))
        report_type, base_model = self.get_report_type_details()

        self.fields['title'] = CharField(initial=data['title'])
        data_attr = split_attr(data)
        self.fields['display_heading'] = BooleanField(required=False,
                                                      widget=RBToggle(),
                                                      label='Display heading')
        if int(data_attr.get('display_heading', 1)) == 1:
            self.fields['display_heading'].initial = True

        if self.django_field is not None and isinstance(self.django_field, DATE_FIELDS):
            self.setup_date_fields(data_attr)
        elif self.django_field is not None and isinstance(self.django_field, NUMBER_FIELDS):
            self.setup_number_fields(data_attr=data_attr,
                                     base_model=base_model,
                                     report_type=report_type)
        elif isinstance(self.col_type_override, CURRENCY_COLUMNS):
            self.setup_currency_fields(data_attr=data_attr)
        elif isinstance(self.col_type_override, LINK_COLUMNS):
            self.setup_link_fields(data_attr=data_attr)
        elif self.col_type_override.annotations is not None:
            self.setup_annotation_fields(data_attr=data_attr)
        elif self.is_mathematical_field(data=data):
            self.setup_standard_mathematical_fields(data=data,
                                                    data_attr=data_attr)
        else:
            self.fields['annotation_label'] = BooleanField(required=False, widget=RBToggle())
            if 'annotation_label' in data_attr and data_attr['annotation_label'] == '1':
                self.fields['annotation_label'].initial = True
        super().setup_modal(*args, **kwargs)

    def setup_annotation_fields(self, data_attr):
        self.fields['show_table_totals'] = BooleanField(required=False,
                                                        widget=RBToggle(),
                                                        label='Show totals')
        if 'show_totals' in data_attr and data_attr['show_totals'] == '1':
            self.fields['show_table_totals'].initial = True

    def setup_currency_fields(self, data_attr):
        self.fields['show_table_totals'] = BooleanField(required=False,
                                                        widget=RBToggle(),
                                                        label='Show totals')
        if 'show_totals' in data_attr and data_attr['show_totals'] == '1':
            self.fields['show_table_totals'].initial = True

    def setup_link_fields(self, data_attr):
        self.fields['link_html'] = CharField(required=False)
        if 'link_html' in data_attr:
            self.fields['link_html'].initial = decode_attribute(data_attr['link_html'])
        self.fields['link_css'] = CharField(required=False)
        if 'link_css' in data_attr:
            self.fields['link_css'].initial = decode_attribute(data_attr['link_css'])
        self.fields['is_icon'] = BooleanField(required=False, widget=RBToggle())
        if 'is_icon' in data_attr and data_attr['is_icon'] == '1':
            self.fields['is_icon'].initial = True

    def setup_date_fields(self, data_attr):
        self.fields['annotations_value'] = ChoiceField(choices=[(0, '-----')] + ANNOTATION_VALUE_CHOICES,
                                                       required=False)
        if 'annotations_value' in data_attr:
            self.fields['annotations_value'].initial = data_attr['annotations_value']
        self.fields['date_format'] = ChoiceField(choices=[(0, '-----')] + DATE_FORMAT_TYPES, required=False)
        if 'date_format' in data_attr:
            self.fields['date_format'].initial = data_attr['date_format']

    def setup_number_fields(self, data_attr, base_model, report_type):
        self.fields['annotations_type'] = ChoiceField(choices=[(0, '-----')] + ANNOTATIONS_CHOICES,
                                                      required=False)
        if 'annotations_type' in data_attr:
            self.fields['annotations_type'].initial = data_attr['annotations_type']

        annotation_column_help_text = 'Not required however useful for mathematical columns.'
        self.fields['annotation_column_id'] = CharField(required=False,
                                                        help_text=annotation_column_help_text)
        if 'annotation_column_id' in data_attr:
            self.fields['annotation_column_id'].initial = decode_attribute(data_attr['annotation_column_id'])

        self.fields['show_table_totals'] = BooleanField(required=False,
                                                        widget=RBToggle(),
                                                        label='Show totals')
        if 'show_totals' in data_attr and data_attr['show_totals'] == '1':
            self.fields['show_table_totals'].initial = True

        self.fields['alignment'] = ChoiceField(choices=ALIGNMENT_CHOICES, required=False)
        if 'alignment' in data_attr:
            self.fields['alignment'].initial = data_attr['alignment']

        self.fields['decimal_places'] = IntegerField()
        self.fields['decimal_places'].initial = int(data_attr.get('decimal_places', 0))
        self.fields['has_filter'] = BooleanField(required=False, widget=RBToggle())

        self.fields['filter'] = CharField(required=False)

        if data_attr.get('has_filter') == '1':
            self.fields['has_filter'].initial = True
            if 'filter' in data_attr:
                self.fields['filter'].initial = decode_attribute(data_attr['filter'])

        self.fields['multiple_columns'] = BooleanField(required=False, widget=RBToggle())

        report_builder_class = get_report_builder_class(model=base_model,
                                                        report_type=report_type)
        fields = []
        self._get_query_builder_foreign_key_fields(base_model=base_model,
                                                   report_builder_class=report_builder_class,
                                                   fields=fields)

        self.fields['multiple_column_field'] = ChoiceField(choices=fields, required=False)

        if data_attr.get('multiple_columns') == '1':
            self.fields['multiple_columns'].initial = True
            self.fields['multiple_column_field'].initial = data_attr.get('multiple_column_field')
            if data_attr.get('append_column_title') == '1':
                self.fields['append_column_title'].initial = True


    def setup_standard_mathematical_fields(self, data, data_attr):

        self.fields['hidden'] = BooleanField(required=False, widget=RBToggle(), label='Hidden')
        if int(data_attr.get('hidden', 0)) == 1:
            self.fields['hidden'].initial = True

        self.fields['column_id'] = CharField(required=False)
        if 'column_id' in data_attr:
            self.fields['column_id'].initial = decode_attribute(data_attr['column_id'])
        if data['field'] in ['rb_division', 'rb_percentage']:
            self.fields['numerator_column'] = CharField(required=False)
            if 'numerator_column' in data_attr:
                self.fields['numerator_column'].initial = decode_attribute(data_attr['numerator_column'])
            self.fields['denominator_column'] = CharField(required=False)
            if 'denominator_column' in data_attr:
                self.fields['denominator_column'].initial = decode_attribute(data_attr['denominator_column'])
        elif data['field'] == 'rb_times':
            self.fields['multiplicand_column'] = CharField(required=False)
            if 'multiplicand_column' in data_attr:
                self.fields['multiplicand_column'].initial = decode_attribute(data_attr['multiplicand_column'])
            self.fields['multiplier_column'] = CharField(required=False)
            if 'multiplier_column' in data_attr:
                self.fields['multiplier_column'].initial = decode_attribute(data_attr['multiplier_column'])
        else:  # add and sub
            self.fields['first_value_column'] = CharField(required=False)
            if 'first_value_column' in data_attr:
                self.fields['first_value_column'].initial = decode_attribute(data_attr['first_value_column'])
            self.fields['second_value_column'] = CharField(required=False)
            if 'second_value_column' in data_attr:
                self.fields['second_value_column'].initial = decode_attribute(data_attr['second_value_column'])
        self.fields['decimal_places'] = IntegerField()
        self.fields['decimal_places'].initial = int(data_attr.get('decimal_places', 0))
        self.fields['show_table_totals'] = BooleanField(required=False,
                                                        widget=RBToggle(),
                                                        label='Show totals')
        if 'show_totals' in data_attr and data_attr['show_totals'] == '1':
            self.fields['show_table_totals'].initial = True

        self.fields['alignment'] = ChoiceField(choices=ALIGNMENT_CHOICES, required=False)
        if 'alignment' in data_attr:
            self.fields['alignment'].initial = data_attr['alignment']

    def save_mathematical_fields(self, data, attributes):
        if self.cleaned_data['hidden']:
            attributes.append('hidden-1')

        alignment = self.cleaned_data.get('alignment')
        if alignment:
            attributes.append(f'alignment-{alignment}')
        if self.cleaned_data['column_id']:
            b64_column_id = encode_attribute(self.cleaned_data['column_id'])
            attributes.append(f'column_id-{b64_column_id}')
        if data['field'] in ['rb_division', 'rb_percentage']:
            if self.cleaned_data['numerator_column']:
                b64_numerator_column = encode_attribute(self.cleaned_data['numerator_column'])
                attributes.append(f'numerator_column-{b64_numerator_column}')
            if self.cleaned_data['denominator_column']:
                b64_denominator_column = encode_attribute(self.cleaned_data['denominator_column'])
                attributes.append(f'denominator_column-{b64_denominator_column}')
        elif data['field'] == 'rb_times':
            if self.cleaned_data['multiplicand_column']:
                b64_multiplicand_column = encode_attribute(self.cleaned_data['multiplicand_column'])
                attributes.append(f'multiplicand_column-{b64_multiplicand_column}')
            if self.cleaned_data['multiplier_column']:
                b64_multiplier_column = encode_attribute(self.cleaned_data['multiplier_column'])
                attributes.append(f'multiplier_column-{b64_multiplier_column}')
        else:
            if self.cleaned_data['first_value_column']:
                b64_first_value_column = encode_attribute(self.cleaned_data['first_value_column'])
                attributes.append(f'first_value_column-{b64_first_value_column}')
            if self.cleaned_data['second_value_column']:
                b64_second_value_column = encode_attribute(self.cleaned_data['second_value_column'])
                attributes.append(f'second_value_column-{b64_second_value_column}')

        attributes.append(f'decimal_places-{self.cleaned_data.get("decimal_places", 0)}')
        if self.cleaned_data['show_table_totals']:
            attributes.append('show_totals-1')

    def save_number_fields(self, attributes):
        alignment = self.cleaned_data.get('alignment')
        if alignment:
            attributes.append(f'alignment-{alignment}')
        if int(self.cleaned_data['annotations_type']) != 0:
            attributes.append(f'annotations_type-{self.cleaned_data["annotations_type"]}')
        if self.cleaned_data['annotation_column_id']:
            b64_annotation_column_id = encode_attribute(self.cleaned_data['annotation_column_id'])
            attributes.append(f'annotation_column_id-{b64_annotation_column_id}')
        if self.cleaned_data['show_table_totals']:
            attributes.append('show_totals-1')
        if self.cleaned_data['decimal_places'] > 0:
            attributes.append(f'decimal_places-{self.cleaned_data["decimal_places"]}')
        if self.cleaned_data['has_filter']:
            attributes.append(f'has_filter-1')
            if self.cleaned_data['filter']:
                b64_filter = encode_attribute(self.cleaned_data['filter'])
                attributes.append(f'filter-{b64_filter}')
            if self.cleaned_data['multiple_columns']:
                attributes.append('multiple_columns-1')
                attributes.append(f'multiple_column_field-{self.cleaned_data["multiple_column_field"]}')

    def save_date_fields(self, attributes):
        if self.cleaned_data['annotations_value']:
            attributes.append(f'annotations_value-{self.cleaned_data["annotations_value"]}')
        if self.cleaned_data['date_format']:
            attributes.append(f'date_format-{self.cleaned_data["date_format"]}')

    def save_currency_fields(self, attributes):
        if self.cleaned_data['show_table_totals']:
            attributes.append('show_totals-1')

    def save_link_fields(self, attributes):
        if self.cleaned_data['link_css']:
            b64_link_css = encode_attribute(self.cleaned_data['link_css'])
            attributes.append(f'link_css-{b64_link_css}')
        if self.cleaned_data['link_html']:
            b64_link_html = encode_attribute(self.cleaned_data['link_html'])
            attributes.append(f'link_html-{b64_link_html}')
        if self.cleaned_data['is_icon'] and self.cleaned_data["is_icon"]:
            attributes.append('is_icon-1')

    def save_annotations_fields(self, attributes):
        if self.cleaned_data['show_table_totals']:
            attributes.append('show_totals-1')

    def get_additional_attributes(self):
        attributes = []
        data = json.loads(base64.b64decode(self.slug['data']))
        self.get_report_type_details()

        if self.cleaned_data['display_heading']:
            attributes.append('display_heading-1')
        else:
            attributes.append('display_heading-0')

        if self.django_field is not None and isinstance(self.django_field, DATE_FIELDS):
            self.save_date_fields(attributes=attributes)
        elif self.django_field is not None and isinstance(self.django_field, NUMBER_FIELDS):
            self.save_number_fields(attributes=attributes)
        elif isinstance(self.col_type_override, CURRENCY_COLUMNS):
            self.save_currency_fields(attributes=attributes)
        elif isinstance(self.col_type_override, LINK_COLUMNS):
            self.save_link_fields(attributes=attributes)
        elif self.col_type_override.annotations is not None:
            self.save_annotations_fields(attributes=attributes)
        elif self.is_mathematical_field(data=data):
            self.save_mathematical_fields(data=data, attributes=attributes)
        else:
            if self.cleaned_data['annotation_label'] and self.cleaned_data["annotation_label"]:
                attributes.append('annotation_label-1')

        if attributes:
            return '-'.join(attributes)
        return None


class TableFieldModal(QueryBuilderModalBaseMixin, FormModal):
    form_class = TableFieldForm
    size = 'xl'
    template_name = 'advanced_report_builder/datatables/fields/modal.html'
    no_header_x = True
    helper_class = HorizontalNoEnterHelper
    update_selection_command = 'update_selection'

    @property
    def modal_title(self):
        data = json.loads(base64.b64decode(self.slug['data']))
        return f'Edit {data["title"]}'

    def form_valid(self, form):
        selector = self.slug['selector']

        _attr = form.get_additional_attributes()
        self.add_command({'function': 'set_attr',
                          'selector': f'#{selector}',
                          'attr': 'data-attr',
                          'val': _attr})

        self.add_command({'function': 'html', 'selector': f'#{selector} span', 'html': form.cleaned_data['title']})
        self.add_command({'function': self.update_selection_command})
        return self.command_response('close')

    def form_setup(self, form, *_args, **_kwargs):

        data = json.loads(base64.b64decode(self.slug['data']))
        report_builder_class, base_model = self.get_report_builder_class(report_type_id=self.slug['report_type_id'])
        django_field, col_type_override, _, _ = self.get_field_details(base_model=base_model,
                                                                       field=data['field'],
                                                                       report_builder_class=report_builder_class)
        if django_field is not None and isinstance(django_field, NUMBER_FIELDS):
            form.add_trigger('annotations_type', 'onchange', [
                {'selector': '#annotations_fields_div', 'values': {'0': 'hide'}, 'default': 'show'}])

            form.add_trigger('has_filter', 'onchange', [
                {'selector': '#filter_fields_div', 'values': {'checked': 'show'}, 'default': 'hide'}])

            form.add_trigger('multiple_columns', 'onchange', [
                {'selector': '#multiple_columns_fields_div', 'values': {'checked': 'show'}, 'default': 'hide'},
            ])

            return ['title',
                    'display_heading',
                    'show_table_totals',
                    'decimal_places',
                    'alignment',
                    'annotations_type',
                    Div('annotation_column_id',
                        FieldEx('has_filter',
                                template='django_modals/fields/label_checkbox.html',
                                field_class='col-6 input-group-sm'),
                        Div(
                            FieldEx('filter',
                                    template='advanced_report_builder/datatables/fields/single_query_builder.html'),
                            FieldEx('multiple_columns',
                                    template='django_modals/fields/label_checkbox.html',
                                    field_class='col-6 input-group-sm'),
                            Div(
                                FieldEx('multiple_column_field'),
                                css_id='multiple_columns_fields_div'),
                            css_id='filter_fields_div'),
                        css_id='annotations_fields_div')
                    ]
        elif isinstance(col_type_override, CURRENCY_COLUMNS):
            return ['title',
                    'show_table_totals']
        elif isinstance(col_type_override, LINK_COLUMNS):
            return ['title',
                    'link_css',
                    'link_html',
                    'is_icon']

    def ajax_get_query_builder_fields(self, **kwargs):
        field_auto_id = kwargs['field_auto_id']

        report_type_id = self.slug['report_type_id']
        query_builder_filters = self.get_query_builder_report_type_field(report_type_id=report_type_id)

        return self.command_response(f'query_builder_{field_auto_id}', data=json.dumps(query_builder_filters))


class TablePivotForm(ChartBaseFieldForm):
    cancel_class = 'btn-secondary modal-cancel'

    def setup_modal(self, *args, **kwargs):
        data = json.loads(base64.b64decode(self.slug['data']))
        self.fields['title'] = CharField(initial=data['title'])
        super().setup_modal(*args, **kwargs)

    def cancel_button(self, css_class=cancel_class, **kwargs):
        commands = [{'function': 'close'}]
        return self.button('Cancel', commands, css_class, **kwargs)


class TablePivotModal(QueryBuilderModalBaseMixin, FormModal):
    form_class = TablePivotForm
    size = 'xl'
    no_header_x = True

    @property
    def modal_title(self):
        data = json.loads(base64.b64decode(self.slug['data']))
        return f'Edit {data["title"]}'

    def form_valid(self, form):
        selector = self.slug['selector']

        self.add_command({'function': 'html', 'selector': f'#{selector} span', 'html': form.cleaned_data['title']})

        self.add_command({'function': 'update_pivot_selection'})
        return self.command_response('close')
